#include "Piece.h"

/**
 * @brief Piece ctor with rotation numbers
 * @param window the TetrisWindow that contains this piece
 * @param numbers numbers used in the rotate function to rotate piece
 */
Piece::Piece(const TetrisWindow* window, const char numbers[])
    : window(window)
    , rotation_numbers(numbers, numbers+24)
    , r(rotation_numbers)
    , move_count(0) {
}

/**
 * @brief Piece ctor without rotation numbers
 * @param window the TetrisWindow that contains this piece
 */
Piece::Piece(const TetrisWindow* window)
    : window(window)
    , r(rotation_numbers)
    , move_count(0) {
}

/**
 * @brief Piece dtor, frees any remaining Blocks that the Piece owns
 */
Piece::~Piece() {
    for(auto b : blocks) {
        delete b;
    }
}

/**
 * @brief Displays the Piece on the screen
 */
void Piece::draw() const {
    for(auto b : blocks) {
        b->draw();
    }
}

/**
 * @brief Moves the Piece in specified direction
 * @param direction -1 is left, 0 is down, 1 is right
 * @retval true move successful
 * @retval false move failed
 */
bool Piece::move(short direction) {
    /* Check if all Blocks of the Piece are movable */
    for(auto b : blocks) {
        if(! b->movable(direction)) {
            /* Report failure */
            return false;
        }
    }
    
    /* Move all Blocks along */
    for(auto b : blocks) {
        b->move(direction);
    }
    
    /* Increase move count if applicable */
    if(direction == 0) {
        move_count++;
    }
    
    /* Report success */
    return true;
}

/**
 * @brief Gets one of the Blocks owned by the Piece,
 *        ownership is transfered, caller is reponsible for Block afterwards
 * @return The last Block stored in the Piece
 */
Block* Piece::get_block() {
    Block* temp = blocks.back();
    blocks.pop_back();
    return temp;
}

/**
 * @brief Check if Piece owns any blocks
 * @retval true the Piece contains at least one Block
 * @retval false the Piece contains no Blocks
 */
bool Piece::has_blocks() const {
    return blocks.size() > 0;
}

/**
 * @brief Rotates a Piece clockwise
 * @details This is a virtual function, as Piece_I and Piece_O have their own unique rotation functions.
 *          Index 1 is the Block to pivot around, the four directions are determined by
 *          the index 0 Block's position from the pivot Block. For each Piece, the coordinates
 *          of where the 0, 2 and 3 index Blocks should end up relative to the pivot Block after 
 *          rotation are stored in 24 item arrays (4 by 6 matrices). In each row, the numbers are 
 *          pairs of (x,y) offsets compared to the pivot Block. For example, the row that puts 
 *          the 3 moving Blocks above the pivot would be:
 *  
 *                  -1, -1,  0, -1,  1, -1
 *                  x1, y1, x2, y2, x3, y3  << All relative to the pivot Block
 * 
 *                          0 2 3           << Blocks above
 *                            1             << Pivot block with index 1
 * 
 *          All the matrices' rows are in order right, left, up, down, as the if-else clauses are below.
 *          These can be found in the IndividualPieces.cpp file.
 * 
 *          For more explaination on how this works, visit the project page at http://zsmb.co/projects/tetris/
 */
void Piece::rotate() {
    int x = blocks[1]->get_x();
    int y = blocks[1]->get_y();
    
    int end_x = blocks[0]->get_x();
    int end_y = blocks[0]->get_y();
    
    if(end_x > x) { /* Piece faces right */
        if(is_clear_to_rotate(x, blocks[1], x+r[0], y+r[1], x+r[2], y+r[3], x+r[4], y+r[5])) {
            blocks[0]->set(x+r[0], y+r[1]);
            blocks[2]->set(x+r[2], y+r[3]);
            blocks[3]->set(x+r[4], y+r[5]);
        }
    }
    else if(end_x < x) { /* Piece faces left */
        if(is_clear_to_rotate(x, blocks[1], x+r[6], y+r[7], x+r[8], y+r[9], x+r[10], y+r[11])) {
            blocks[0]->set(x+r[6], y+r[7]);
            blocks[2]->set(x+r[8], y+r[9]);
            blocks[3]->set(x+r[10], y+r[11]);
        }
    }
    else if(end_y < y) { /* Piece faces up */
        if(is_clear_to_rotate(x, blocks[1], x+r[12], y+r[13], x+r[14], y+r[15], x+r[16], y+r[17])) {
            blocks[0]->set(x+r[12], y+r[13]);
            blocks[2]->set(x+r[14], y+r[15]);
            blocks[3]->set(x+r[16], y+r[17]);
        }
    }
    else { /* Piece faces down */
        if(is_clear_to_rotate(x, blocks[1], x+r[18], y+r[19], x+r[20], y+r[21], x+r[22], y+r[23])) {
            blocks[0]->set(x+r[18], y+r[19]);
            blocks[2]->set(x+r[20], y+r[21]);
            blocks[3]->set(x+r[22], y+r[23]);
        }
    }
}

/**
 * @brief Checks if positions specified around a Piece are clear for rotation
 * @details In case the positions are unavailable, tries moving the Piece left and right
 *          a maximum of 2 units and checks if it could be rotated in those positions - 
 *          has the ability to move the pivot block as well as modify the x parameter,
 *          the latter of which is to be used by the caller when moving the Blocks of the
 *          Piece around
 * @param x
 * @param pivot_block - the Block that the Piece is being rotated around
 * @param x1 x coordinate of first position
 * @param y1 y coordinate of first position
 * @param x2 x coordinate of second position
 * @param y2 y coordinate of second position
 * @param x3 x coordinate of third position
 * @param y3 y coordinate of third position
 * @retval true a position available for rotation has been found
 * @retval false the piece can not be rotated
 */
bool Piece::is_clear_to_rotate(int& x, Block* pivot_block, int x1, int y1, int x2, int y2, int x3, int y3) const {
    /* Try to rotate in place, try moving the piece left and right
     * a bit if that's not possible */
    int offsets[] = {0, -1, 1, -2, 2};
    
    for(int o : offsets) {
        if(is_clear(x1+o, y1, x2+o, y2, x3+o, y3)) {
            /* Most of the time we rotate in place, only change things if 
             * it's really necessary */
            if(o != 0) {
                /* Move pieces by offset */
                x += o;
                pivot_block->set(x, pivot_block->get_y());
            }
            /* Report success */
            return true;
        }
    }
    
    /* Report failure */
    return false;
}

/**
 * @brief Checks if three positions in the TetrisWindow are available
 * @param x1 x coordinate of first position
 * @param y1 y coordinate of first position
 * @param x2 x coordinate of second position
 * @param y2 y coordinate of second position
 * @param x3 x coordinate of third position
 * @param y3 y coordinate of third position
 * @retval true all the specified positions are available
 * @retval false the specified positions are not available
 */
bool Piece::is_clear(int x1, int y1, int x2, int y2, int x3, int y3) const {
    return window->is_available(x1, y1) &&
           window->is_available(x2, y2) &&
           window->is_available(x3, y3);
}

/**
 * @brief Getter function for move count
 * @return the number of times the block moved downwards
 */
unsigned Piece::get_move_count() const {
    return move_count;
}
